#include <Bounce2.h>
#include <TimerOne.h>
#include <ClickEncoder.h>
#include <LiquidCrystal.h>

//#define tic 0x927C0
#define tic 0x493E0

#define EN_PIN   4 // Пин включения драйвера
#define STEP_PIN  3 // Пин сигнала шага драйвера (CLC, STEP)
#define DIR_PIN  2 //Пин направления вращения (CW, DIR

#define RS_PIN 5 // Пин сигнала RES дисплея
#define RW_PIN 6 // Пин сигнала RW дисплея
#define E_PIN 12 // Пин сигнала En дисплея

#define D4_PIN 10 // Пин полубайта данных дисплея
#define D5_PIN 11 // Пин полубайта данных дисплея
#define D6_PIN 8 // Пин полубайта данных дисплея
#define D7_PIN 9 // Пин полубайта данных дисплея

#define LCD_ROW 2 // Строк дисплея
#define LCD_COL 16 // Символов в строке

#define ENCODER_A A3
#define ENCODER_B A2
#define ENCODER_CLICK A1

#define STOP_PIN A5

LiquidCrystal lcd(RS_PIN, RW_PIN, E_PIN, D4_PIN, D5_PIN, D6_PIN, D7_PIN); //Конфигурируем дисплей

ClickEncoder *encoder;
int16_t last, value;

void timerIsr() {
	encoder->service();
}


volatile uint16_t count;
volatile unsigned long speed;
volatile bool run;
bool start, dir, timer_run, info;
float minRpm, maxRpm;
uint8_t mode;

float rpm;
unsigned long key_time[5];

String serialCommand = "";
boolean commandReceived = false;
Bounce stopPin = Bounce();

boolean globalStop = false;

void setup()
{
	Serial.begin(9600);
	serialCommand.reserve(200);

	lcd.begin(LCD_COL, LCD_ROW);

	pinMode(STEP_PIN, OUTPUT);
	pinMode(DIR_PIN, OUTPUT);
	pinMode(EN_PIN, OUTPUT);

	pinMode(STOP_PIN, INPUT_PULLUP);
	digitalWrite(STOP_PIN, HIGH);
	stopPin.attach(STOP_PIN);
	stopPin.interval(5);

	encoder = new ClickEncoder(ENCODER_A, ENCODER_B, ENCODER_CLICK, 2);
	Timer1.initialize(1000);
	Timer1.attachInterrupt(timerIsr);

	maxRpm = 300;
	minRpm = 0.1;
	rpm = 100;
	dir = false;
	speed = tic / rpm;
	run = false;
	start = false;

	TCCR2A = (1 << WGM21);
	TCCR2B = (1 << CS20);
	OCR2A = 0x80;
	TIMSK2 = (1 << OCIE2A);

	sei();
}

void loop()
{
	stopPin.update();
	disp();

	if (start) {
		accelerate();
	}

	processSerial();

	action();
	calculate();
	set_pins();
}

void action()
{
	int delta = encoder->getValue();

	if (delta != 0)
	{
		if (rpm >= 30) {
			rpm = (int)rpm;
			rpm += delta;
		}
		else {
			rpm += delta * 0.1;
		}

		if (rpm > maxRpm)
		{
			rpm = maxRpm;
		}
		if (rpm < minRpm)
		{
			rpm = minRpm;
		}
	}

	ClickEncoder::Button b = encoder->getButton();
	if (b != ClickEncoder::Open)
	{
		switch (b)
		{
		case ClickEncoder::Clicked:
			if (!run)
			{
				doStart();
			}
			else {
				doStop();
			}
			break;
		case ClickEncoder::DoubleClicked:
			change_dir();
			break;
		}
	}

	if (stopPin.read() == 0)
	{
		globalStop = true;
	}

	if (stopPin.read() == 1)
	{
		globalStop = false;
	}

}

void doStop()
{
	if (run)
	{
		run = false;
	}
}

void doStart()
{
	if (!run)
	{
		start = true;
	}
}

void accelerate()
{
	run = true;
	set_pins();
	int i;
	for (i = 1; i<rpm; i++)
	{
		speed = tic / (i);
		delay(3);
	}
	start = false;
}

void decelerate()
{
	int i;
	for (i = rpm; i > 10; i--)
	{
		speed = tic / i;
		delay(1);
	}
	run = false;
}


void disp()
{
	if (millis() % 330)
	{
		char rpm_str[7] = "      ";
		dtostrf(rpm, 3, 1, rpm_str);

		lcd.setCursor(0, 0);
		lcd.print(rpm_str);
		lcd.setCursor(5, 0);
		lcd.print(dir ? '>' : '<');

		lcd.setCursor(0, 1);
		lcd.print(run ? "ON " : "OFF");
	}
}

void change_dir()
{
	if (run)
	{
		decelerate();
		dir = !dir;
		start = true;
	}
	else
	{
		dir = !dir;
	}
}

void set_pins()
{
	digitalWrite(EN_PIN, !globalStop && run ? LOW : HIGH);
	digitalWrite(DIR_PIN, dir ? LOW : HIGH);
}

void calculate()
{
	speed = tic / rpm;
}

void processSerial()
{
	if (commandReceived) {
		if (serialCommand == "AT+START")
		{
			doStart();
			Serial.println("OK");
		}
		else if (serialCommand == "AT+STOP")
		{
			doStop();
			Serial.println("OK");
		}
		else if (serialCommand == "AT")
		{
			Serial.println("OK");
		}
		else if (serialCommand.startsWith("AT+RPM="))
		{
			float newRpm = serialCommand.substring(7).toFloat();

			if (newRpm > maxRpm || newRpm < minRpm)
			{
				Serial.println("ERROR");
				Serial.println("Invalid RPM.");
			}
			else
			{
				rpm = newRpm;
				Serial.println("OK");
			}
		}
		else if (serialCommand == "AT+RPM")
		{
			Serial.println(rpm);
			Serial.println("OK");
		}
		else
		{
			Serial.println("ERROR");
			Serial.println("Unknown command.");
		}

		serialCommand = "";
		commandReceived = false;
	}
}

void serialEvent() {
	while (Serial.available())
	{
		char inChar = (char)Serial.read();
		if (inChar == '\n') {
			commandReceived = true;
		}
		else {
			serialCommand += inChar;
		}
	}
}

ISR(TIMER2_COMPA_vect)
{
	if (count > 0)
	{
		count--;
		OCR2A = 0x80;
	}
	else
	{
		digitalWrite(STEP_PIN, HIGH);
		count = speed >> 7;
		OCR2A = speed % 0x80;
		digitalWrite(STEP_PIN, LOW);
	}
}


